<?php

/**
 * RoundCube iframe_external plugin.
 * Allows login and run the webmail client from an external iframe.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
 * Based on Clevercube External authentication plugin by Günter Kits, 2013.
 *
 * @version 0.2
 * @author Zoltán Kovács <kzoli@kzoli.hu>
 * @license http://www.gnu.org/licenses/gpl-3.0.html GPL-3.0
 *
 */

class iframe_external extends rcube_plugin
{
    private $rc;
    private $config;
    private $origin;

    public function init()
    {
        $this->rc = rcmail::get_instance();
        $this->config = $this->rc->config;

        // Load user's local configuration.
        $this->load_config('config/config.inc.php');

	// If no allowed origin for external iframing, there is nothing to do.
	if (is_array($this->config->get('iframe_external_allowed_origins'))) {

    	    // Registering hooks.
	    $this->add_hook('config_get', array($this, 'config_get'));
    	    $this->add_hook('authenticate', array($this, 'authenticate'));
    	    $this->add_hook('render_page', array($this, 'render_page'));
	    // Registering a shutdown function.
    	    $this->rc->add_shutdown_function(array($this, 'shutdown'));

	    // Save the iframe origin at initialisation, if any and if it is allowed.
	    // We guess the origin from the SESSION, or from a GET or POST parameter,
	    // or from the HTTP_REFERER server variable.
	    // We will use the http(s)://FULLY.QUALIFIED.HOSTNAME format, without trailing slash.
	    if (!empty($_SESSION['_origin']))
		$referrer = implode('/',array_slice(explode('/', urldecode($_SESSION['_origin'])),0,3));
	    else if (!empty($_GET['_origin']))
		$referrer = implode('/',array_slice(explode('/', urldecode($_GET['_origin'])),0,3));
	    else if (!empty($_POST['_origin']))
		$referrer = implode('/',array_slice(explode('/', urldecode($_POST['_origin'])),0,3));
	    else
		$referrer = implode('/',array_slice(explode('/', $_SERVER['HTTP_REFERER']),0,3));
	    // It is valid origin when isn't empty and was listed in plugin's configuraion.
	    if (!empty($referrer) &&
		in_array($referrer, $this->config->get('iframe_external_allowed_origins',array('n/a')))) {
		// Save the origin within this object.
		$this->origin = $referrer;
		// Save in SESSION also for JavaScript page reloads or redirects.
		$_SESSION['_origin'] = $referrer;
	    }
	}
    }


 /**
  * This hook disables cross site request forgery check (http://en.wikipedia.org/wiki/Cross-site_request_forgery)
  * and cookie check for external login attemtps.
  *
  * @todo	Consider the safety risks.
  *
  * @param array	args	parameter array
  *   string	host		hostname
  *   string	user		the username to authenticate
  *   string	pass		a plain text password
  *   booean	cookiecheck	false to disable cookie check
  *   boolean	valid		true to disable cross site request forgery (CSRF) check
  *   boolean	abort		true to skip the imap login and fail
  *   string	error		if using abort, also set error to any string (else you will get a storage error)
  *
  * @return array	args	maybe modified parameter array
  */
    public function authenticate($args)
    {
        if ($this->rc->action == 'login' && $this->rc->task == 'login' &&
	    !empty($this->origin) &&
	    !empty($_POST['_user']) && !empty($_POST['_pass']))
        {
            $args['cookiecheck'] = false;	// disable cookie check
            $args['valid'] = true;		// disable CSRF
        }
        return $args;
    }


 /**
  * This hook changes the $config['x_frame_options'] value to the allowed origin,
  * saved at initialisation. This function mimics editing the config file itself,
  * but for this session only.
  *
  * @todo	Handling the rcmail::config->all() request with '*' in $args['name'].
  *
  * @param array	args	parameter array
  *   name	string	name of the config option
  *   default	string	default value of the config option
  *   result	string	value of the config option
  *
  * @return array	args	maybe modified parameter array
  */
    public function config_get($args)
    {
	if ($args['name'] == 'x_frame_options' && !empty($this->origin))
	{
	    // Content of an appropriate X-Frame_Options header.
	    $args['result'] = 'allow-from ' . $this->origin;
	}
	return $args;
    }

  /**
   * This hook sends an appropriate X-Frame-Options HTTP header, if an allowed origin was saved
   * at initialisation. Additionally, passes an env option to the client side, to inform the
   * client's clickjack protection JavaScript. Finally, hacks the re-login form, which appears
   * after user's logout within iframe, to resend the lost origin from the client side.
   *
   * @param string	content		The HTML source.
   *
   * @return string	content		The HTML source (maybe modified).
   */
    public function render_page($content)
    {
	// We have a saved origin, and it is possible to append a HTTP header.
        if (!empty($this->origin) && !headers_sent()) {
	    header('X-Frame-Options: allow-from ' . $this->origin);
	    // Without this hack, IE will ignore cookies from iframe.
	    header('P3P: CP="Dummy"');
	    // Pass to the client side (the rcmail JavaScript) also.
	    $this->rc->output->set_env('x_frame_options', 'allow-from ' . $this->origin); }
	else if (!empty($_SESSION['_origin']) && !headers_sent()) {
	    header('X-Frame-Options: allow-from ' . $_SESSION['_origin']);
	    // Without this hack, IE will ignore cookies from iframe.
	    header('P3P: CP="Dummy"');
	    // Pass to the client side (the rcmail JavaScript) also.
	    $this->rc->output->set_env('x_frame_options', 'allow-from ' . $S_SESSION['_origin']);
	}

	// The re-rendered login form (appears after logout from Roundcube running in the iframe)
	// is a special case; we need inject a JavaScript which will include the origin from
	// client-side Roundcube environment into the POST-ed credentials.
	if ($this->rc->task == 'login') {
	    $this->include_script('js/iframe_external.js');
	}
	return $content;
    }


 /**
  * This pseudo-hook handles the Roundcube internal redirects performed by Location: header.
  * Invoked at shutdown (typically by exit() or die() call) via PHP register_shutdown_function.
  * Appends the URL-encoded origin (if any) to the GET line.
  */
    public function shutdown()
    {
	// We have a saved origin, and it is possible to append a HTTP header.
        if (!empty($this->origin) && !headers_sent()) {
	    // Search the Location: header (if any).
	    foreach (headers_list() as $header) {
		if (strtolower(substr($header,0,9)) == 'location:') {
		    // Appand the URL-encoded origin.
		    header($header . '&_origin=' . urlencode($this->origin));
		    break;
		}
	    }
	    // Maybe overzealous here :-)
    	    header('X-Frame-Options: allow-from ' . $this->origin);
	    // Without this hack, IE will ignore cookies from iframe.
	    header('P3P: CP="Dummy"');
        }
	return;
    }

}
?>
